/*************************************************************************
 *
 *   Program:  MPU-9150_Processor			Version: 1.00
 *
 *   Author:  SS							Date:  November 27, 2014
 *
 *   This files contains a set of functions to configure and access the Invensence MPU-9150 9 DOF sensor.
 *
 *   SOC Robotics, Inc is not liable in anyway for the use of this software and assumes no responsibility 
 *   for it's use.  No warranties of any kind are provided.
 *
 *   Copyright 2014, SOC Robotics, Inc.  
 *   There maybe code from other parties in this application  If we don't mention them by name thank you 
 *   and if we should please let us know and we will include your name and attribution.
 *
 *************************************************************************/
//#include "hw_config.h"

#include "stdint.h"
#include "stdbool.h"
#include "stdio.h"
#include "math.h"
#include "string.h"
#include "stddef.h"
#include "stdlib.h"
#include "ctype.h"

#include "NRF24L01.h"
#include "PicoM_Defines.h"
#include "PicoM_Externals.h"
#include "colors.h"

#include "MadgwickAHRS.h"

/**************************************************************************
	IMU Structure
**************************************************************************/

volatile struct imudata {
	short int ai[3];
	short int gi[3];
	short int mi[3];
	short int temp;
	float magcor[3];
	float af[3];
	float gf[3];
	float mf[3];
	short int tempi;
	short int rolli;
	short int pitchi;
	short int yawi;
	short int pressi;
	float tempf;
	float roll;
	float pitch;
	float yaw;
	float press;
	char mget;
} imu;


/***************************************************
	MPU-9150 Configuration Defines
****************************************************/

// Slave addresses
#define  MPU9150_ADDR       0x68        // TWI slave bus address
//#define  MPU9150_ADDR		0x69        // Alternate address
#define  MPU9150_MAG_ADDRESS 0x0C

// Register names according to the datasheet.
// According to the InvenSense document
// "MPU-9150 Register Map and Descriptions Revision 4.0",
#define MPU9150_SELF_TEST_X        0x0D   // R/W
#define MPU9150_SELF_TEST_Y        0x0E   // R/W
#define MPU9150_SELF_TEST_Z        0x0F   // R/W
#define MPU9150_SELF_TEST_A        0x10   // R/W
#define MPU9150_SMPLRT_DIV         0x19   // R/W
#define MPU9150_CONFIG             0x1A   // R/W
#define MPU9150_GYRO_CONFIG        0x1B   // R/W
#define MPU9150_ACCEL_CONFIG       0x1C   // R/W
#define MPU9150_FF_THR             0x1D   // R/W
#define MPU9150_FF_DUR             0x1E   // R/W
#define MPU9150_MOT_THR            0x1F   // R/W
#define MPU9150_MOT_DUR            0x20   // R/W
#define MPU9150_ZRMOT_THR          0x21   // R/W
#define MPU9150_ZRMOT_DUR          0x22   // R/W
#define MPU9150_FIFO_EN            0x23   // R/W
#define MPU9150_I2C_MST_CTRL       0x24   // R/W
#define MPU9150_I2C_SLV0_ADDR      0x25   // R/W
#define MPU9150_I2C_SLV0_REG       0x26   // R/W
#define MPU9150_I2C_SLV0_CTRL      0x27   // R/W
#define MPU9150_I2C_SLV1_ADDR      0x28   // R/W
#define MPU9150_I2C_SLV1_REG       0x29   // R/W
#define MPU9150_I2C_SLV1_CTRL      0x2A   // R/W
#define MPU9150_I2C_SLV2_ADDR      0x2B   // R/W
#define MPU9150_I2C_SLV2_REG       0x2C   // R/W
#define MPU9150_I2C_SLV2_CTRL      0x2D   // R/W
#define MPU9150_I2C_SLV3_ADDR      0x2E   // R/W
#define MPU9150_I2C_SLV3_REG       0x2F   // R/W
#define MPU9150_I2C_SLV3_CTRL      0x30   // R/W
#define MPU9150_I2C_SLV4_ADDR      0x31   // R/W
#define MPU9150_I2C_SLV4_REG       0x32   // R/W
#define MPU9150_I2C_SLV4_DO        0x33   // R/W
#define MPU9150_I2C_SLV4_CTRL      0x34   // R/W
#define MPU9150_I2C_SLV4_DI        0x35   // R
#define MPU9150_I2C_MST_STATUS     0x36   // R
#define MPU9150_INT_PIN_CFG        0x37   // R/W
#define MPU9150_INT_ENABLE         0x38   // R/W
#define MPU9150_INT_STATUS         0x3A   // R
#define MPU9150_ACCEL_XOUT_H       0x3B   // R
#define MPU9150_ACCEL_XOUT_L       0x3C   // R
#define MPU9150_ACCEL_YOUT_H       0x3D   // R
#define MPU9150_ACCEL_YOUT_L       0x3E   // R
#define MPU9150_ACCEL_ZOUT_H       0x3F   // R
#define MPU9150_ACCEL_ZOUT_L       0x40   // R
#define MPU9150_TEMP_OUT_H         0x41   // R
#define MPU9150_TEMP_OUT_L         0x42   // R
#define MPU9150_GYRO_XOUT_H        0x43   // R
#define MPU9150_GYRO_XOUT_L        0x44   // R
#define MPU9150_GYRO_YOUT_H        0x45   // R
#define MPU9150_GYRO_YOUT_L        0x46   // R
#define MPU9150_GYRO_ZOUT_H        0x47   // R
#define MPU9150_GYRO_ZOUT_L        0x48   // R
#define MPU9150_EXT_SENS_DATA_00   0x49   // R
#define MPU9150_EXT_SENS_DATA_01   0x4A   // R
#define MPU9150_EXT_SENS_DATA_02   0x4B   // R
#define MPU9150_EXT_SENS_DATA_03   0x4C   // R
#define MPU9150_EXT_SENS_DATA_04   0x4D   // R
#define MPU9150_EXT_SENS_DATA_05   0x4E   // R
#define MPU9150_EXT_SENS_DATA_06   0x4F   // R
#define MPU9150_EXT_SENS_DATA_07   0x50   // R
#define MPU9150_EXT_SENS_DATA_08   0x51   // R
#define MPU9150_EXT_SENS_DATA_09   0x52   // R
#define MPU9150_EXT_SENS_DATA_10   0x53   // R
#define MPU9150_EXT_SENS_DATA_11   0x54   // R
#define MPU9150_EXT_SENS_DATA_12   0x55   // R
#define MPU9150_EXT_SENS_DATA_13   0x56   // R
#define MPU9150_EXT_SENS_DATA_14   0x57   // R
#define MPU9150_EXT_SENS_DATA_15   0x58   // R
#define MPU9150_EXT_SENS_DATA_16   0x59   // R
#define MPU9150_EXT_SENS_DATA_17   0x5A   // R
#define MPU9150_EXT_SENS_DATA_18   0x5B   // R
#define MPU9150_EXT_SENS_DATA_19   0x5C   // R
#define MPU9150_EXT_SENS_DATA_20   0x5D   // R
#define MPU9150_EXT_SENS_DATA_21   0x5E   // R
#define MPU9150_EXT_SENS_DATA_22   0x5F   // R
#define MPU9150_EXT_SENS_DATA_23   0x60   // R
#define MPU9150_MOT_DETECT_STATUS  0x61   // R
#define MPU9150_I2C_SLV0_DO        0x63   // R/W
#define MPU9150_I2C_SLV1_DO        0x64   // R/W
#define MPU9150_I2C_SLV2_DO        0x65   // R/W
#define MPU9150_I2C_SLV3_DO        0x66   // R/W
#define MPU9150_I2C_MST_DELAY_CTRL 0x67   // R/W
#define MPU9150_SIGNAL_PATH_RESET  0x68   // R/W
#define MPU9150_MOT_DETECT_CTRL    0x69   // R/W
#define MPU9150_USER_CTRL          0x6A   // R/W
#define MPU9150_PWR_MGMT_1         0x6B   // R/W
#define MPU9150_PWR_MGMT_2         0x6C   // R/W
#define MPU9150_FIFO_COUNTH        0x72   // R/W
#define MPU9150_FIFO_COUNTL        0x73   // R/W
#define MPU9150_FIFO_R_W           0x74   // R/W
#define MPU9150_WHO_AM_I           0x75   // R

//  MPU9150 Compass
#define MPU9150_CMPS_XOUT_L        0x4A   // R
#define MPU9150_CMPS_XOUT_H        0x4B   // R
#define MPU9150_CMPS_YOUT_L        0x4C   // R
#define MPU9150_CMPS_YOUT_H        0x4D   // R
#define MPU9150_CMPS_ZOUT_L        0x4E   // R
#define MPU9150_CMPS_ZOUT_H        0x4F   // R

#define MPU9150_ASAX			   0x10
#define MPU9150_ASAY			   0x11
#define MPU9150_ASAZ			   0x12

// I2C address 0x69 could be 0x68 depends on your wiring.
// Declared at variable so functions can switch from accel/gyro to mag
volatile unsigned char  MPU9150_I2C_ADDRESS	=	0x68;

/***************************************************
	MPU-9150 TWI Functions
****************************************************/

volatile int MPU9150_ReadSensor(volatile int addrL, volatile int addrH)
{
	union {
		volatile short int intvalue;
		volatile unsigned char charbuf[2];
	} returnvalue;

	messageBuf[0] = (unsigned char) addrL;
	//TWI_Write(MPU9150_I2C_ADDRESS,messageBuf,1);
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C2,MPU9150_I2C_ADDRESS,1);
#else
	I2C_WriteBytes(I2C1,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C1,MPU9150_I2C_ADDRESS,1);
#endif
	returnvalue.charbuf[0] = messageBuf[0];
	Delaymsec(2);
	
	messageBuf[0] = (unsigned char) addrH;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C2,MPU9150_I2C_ADDRESS,1);
#else
	I2C_WriteBytes(I2C1,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C1,MPU9150_I2C_ADDRESS,1);
#endif
	returnvalue.charbuf[1] = messageBuf[0];
	Delaymsec(2);

	return returnvalue.intvalue;
}

volatile unsigned char MPU9150_ReadSensor2(volatile unsigned char addr)
{
	messageBuf[0] = addr;
	//printf("addr=%d  messgeBuf[0]=%x\n\r",addr,messageBuf[0]);
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_I2C_ADDRESS,1);
#else
	I2C_WriteBytes(I2C1,MPU9150_I2C_ADDRESS,1);
#endif	messageBuf[0] = 0;
#ifdef NANOMTARGET
	I2C_ReadBytes(I2C2,MPU9150_I2C_ADDRESS,1);
#else
	I2C_ReadBytes(I2C1,MPU9150_I2C_ADDRESS,1);
#endif
	//printf("Return byte=%x\n\r",messageBuf[0]);

	return messageBuf[0];
}

volatile int MPU9150_WriteSensor(volatile int addr,volatile int data)
{
	messageBuf[0] = (unsigned char) addr;
	messageBuf[1] = (unsigned char) data;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_I2C_ADDRESS,2);
#else
	I2C_WriteBytes(I2C1,MPU9150_I2C_ADDRESS,2);
#endif
	delay_ms(2);
	return 1;
}

// http://pansenti.wordpress.com/2013/03/26/pansentis-invensense-mpu-9150-software-for-arduino-is-now-on-github/
// Thank you to pansenti for setup code.
void MPU9150_SetupCompass(void)
{
	MPU9150_I2C_ADDRESS = 0x0B;      // change Address to Compass

	MPU9150_WriteSensor(0x0A, 0x00); // PowerDownMode
	MPU9150_WriteSensor(0x0A, 0x0F); // SelfTest
	MPU9150_WriteSensor(0x0A, 0x00); // PowerDownMode

	MPU9150_I2C_ADDRESS = 0x68;      // change Address to MPU

	MPU9150_WriteSensor(0x24, 0x40); // Wait for Data at Slave0
	MPU9150_WriteSensor(0x25, 0x8C); // Set i2c address at slave0 at 0x0C
	MPU9150_WriteSensor(0x26, 0x02); // Set where reading at slave 0 starts
	MPU9150_WriteSensor(0x27, 0x88); // set offset at start reading and enable
	MPU9150_WriteSensor(0x28, 0x0C); // set i2c address at slv1 at 0x0C
	MPU9150_WriteSensor(0x29, 0x0A); // Set where reading at slave 1 starts
	MPU9150_WriteSensor(0x2A, 0x81); // Enable at set length to 1
	MPU9150_WriteSensor(0x64, 0x01); // overvride register
	MPU9150_WriteSensor(0x67, 0x03); // set delay rate
	MPU9150_WriteSensor(0x01, 0x80);

	MPU9150_WriteSensor(0x34, 0x04); // set i2c slv4 delay
	MPU9150_WriteSensor(0x64, 0x00); // override register
	MPU9150_WriteSensor(0x6A, 0x00); // clear usr setting
	MPU9150_WriteSensor(0x64, 0x01); // override register
	MPU9150_WriteSensor(0x6A, 0x20); // enable master i2c mode
	MPU9150_WriteSensor(0x34, 0x13); // disable slv4
}

void MPU9150_Setup(void)
{
	// Reset MPU9150
	MPU9150_WriteSensor(MPU9150_PWR_MGMT_1,0x80);
	delay_ms(1);
	
	// Clear the 'sleep' bit to start the sensor.
	MPU9150_WriteSensor(MPU9150_PWR_MGMT_1, 1);

	MPU9150_WriteSensor(MPU9150_SMPLRT_DIV,0x01);   // 25  - SMPLRT_DIV = 1 => 1khz/(1+1) = 500hz sample rate
	MPU9150_WriteSensor(MPU9150_CONFIG, 0x00);   // 26  - Set DLPF_CONFIG to 4 for 20Hz bandwidth
	MPU9150_WriteSensor(MPU9150_GYRO_CONFIG,  0x18);  //  Full range 2000d/sec
	//MPU9150_writeSensor(MPU9150_ACCEL_CONFIG, 0x10);   // 8G
	MPU9150_WriteSensor(MPU9150_ACCEL_CONFIG, 0x18);   // 16G	
	MPU9150_WriteSensor(MPU9150_USER_CTRL, 0);   
	//MPU9150_writeSensor(MPU9150_INT_PIN_CFG,  %00110010);  // 0x32 
	MPU9150_WriteSensor(MPU9150_INT_PIN_CFG,0x02);
	MPU9150_WriteSensor(MPU9150_INT_ENABLE,0x01); 

	//MPU9150_setupCompass();
}

//  Compass sensitivity is 0.3uT, 13bit A/D (4096), +-1200uT
//  Earth's magnetic field varies from 25 to 65 uT
// Local Vancouver inclination and declination -
// LOCALINCLINATION 70.367
// LOCALDECLINATION 16.777

// Local field strength
// LOCALHNT  18.379
// LOCALVNT  51.579
// LOCALFIELDSTRENGTH 54.699
// LOCALFIELDX  17.596
// LOCALFIELDY  5.305
void MPU9150_Read_Mag(void)
{
	union {
		volatile short int mi;
		volatile unsigned char charbuf[2];
	} mag;

	messageBuf[0] = 0x0A;
	messageBuf[1] = 0x01;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_MAG_ADDRESS,2);
#else
	I2C_WriteBytes(I2C1,MPU9150_MAG_ADDRESS,2);
#endif
	delay_ms(10);

	messageBuf[0] = 0x03;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_MAG_ADDRESS,1);
	I2C_ReadBytes(I2C2,MPU9150_MAG_ADDRESS,6);
#else
	I2C_WriteBytes(I2C1,MPU9150_MAG_ADDRESS,1);
	I2C_ReadBytes(I2C1,MPU9150_MAG_ADDRESS,6);
#endif

	mag.charbuf[0] = messageBuf[0];
	mag.charbuf[1] = messageBuf[1];
	imu.mi[0] = mag.mi;
	imu.mf[0] = (float)mag.mi*(1200.0/4095.0)*imu.magcor[0];
	mag.charbuf[0] = messageBuf[2];
	mag.charbuf[1] = messageBuf[3];
	imu.mi[1] = mag.mi;
	imu.mf[1] = (float)mag.mi*(1200.0/4095.0)*imu.magcor[1];
	mag.charbuf[0] = messageBuf[4];
	mag.charbuf[1] = messageBuf[5];
	imu.mi[2] = mag.mi;
	imu.mf[2] = (float)mag.mi*(1200.0/4095.0)*imu.magcor[2];
	
	//MPU9150_writeSensor(MPU9150_INT_PIN_CFG,  0x00);				
}

// Retrieve the magnetic sensitivity values
void MPU9150_Read_Mag_Sense(void)
{
	unsigned char correctionvalue;
	int i;

	messageBuf[0] = 0x0A;
	messageBuf[1] = 0x0F;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_MAG_ADDRESS,2);
#else
	I2C_WriteBytes(I2C1,MPU9150_MAG_ADDRESS,2);
#endif

	messageBuf[0] = MPU9150_ASAX;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_MAG_ADDRESS,1);
	I2C_ReadBytes(I2C2,MPU9150_MAG_ADDRESS,3);
#else
	I2C_WriteBytes(I2C1,MPU9150_MAG_ADDRESS,1);
	I2C_ReadBytes(I2C1,MPU9150_MAG_ADDRESS,3);
#endif

	for(i=0; i<3; i++) {
	correctionvalue = messageBuf[i];
		imu.magcor[i] = (float)(correctionvalue - 128)*0.5/128.0 + 1.0;
		}
}

void MPU9150_Read_Accel(void)
{
	union {
		volatile short int ai;
		volatile unsigned char charbuf[2];
	} accel;

	messageBuf[0] = (unsigned char) MPU9150_ACCEL_XOUT_H;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C2,MPU9150_I2C_ADDRESS,6);
#else
	I2C_WriteBytes(I2C1,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C1,MPU9150_I2C_ADDRESS,6);
#endif
	accel.charbuf[1] = messageBuf[0];
	accel.charbuf[0] = messageBuf[1];
	imu.ai[0] = accel.ai;
	imu.af[0] = (float)accel.ai/2048.0;
	accel.charbuf[1] = messageBuf[2];
	accel.charbuf[0] = messageBuf[3];
	imu.ai[1] = accel.ai;
	imu.af[1] = (float)accel.ai/2048.00;
	accel.charbuf[1] = messageBuf[4];
	accel.charbuf[0] = messageBuf[5];
	imu.ai[2] = accel.ai;
	imu.af[2] = (float)accel.ai/2048.00;
}

void MPU9150_Read_Gyro(void)
{
	union {
		volatile short int gi;
		volatile unsigned char charbuf[2];
	} gyro;

	messageBuf[0] = (unsigned char) MPU9150_GYRO_XOUT_H;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C2,MPU9150_I2C_ADDRESS,6);
#else
	I2C_WriteBytes(I2C1,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C1,MPU9150_I2C_ADDRESS,6);
#endif
	gyro.charbuf[1] = messageBuf[0];
	gyro.charbuf[0] = messageBuf[1];
	imu.gi[0] = gyro.gi;
	imu.gf[0] = (float)gyro.gi/16.384;  //  2000/32768
	gyro.charbuf[1] = messageBuf[2];
	gyro.charbuf[0] = messageBuf[3];
	imu.gi[1] = gyro.gi;
	imu.gf[1] = (float) gyro.gi/16.384;
	gyro.charbuf[1] = messageBuf[4];
	gyro.charbuf[0] = messageBuf[5];
	imu.gi[2] = gyro.gi;
	imu.gf[2] = (float)gyro.gi/16.384;
}

void MPU9150_Read_Temp(void)
{
	union {
		volatile short int ti;
		volatile unsigned char charbuf[2];
	} temp;

	messageBuf[0] = (unsigned char) MPU9150_TEMP_OUT_H;
#ifdef NANOMTARGET
	I2C_WriteBytes(I2C2,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C2,MPU9150_I2C_ADDRESS,2);
#else
	I2C_WriteBytes(I2C1,MPU9150_I2C_ADDRESS,1);
	I2C_ReadBytes(I2C1,MPU9150_I2C_ADDRESS,2);
#endif
	temp.charbuf[1] = messageBuf[0];
	temp.charbuf[0] = messageBuf[1];
	imu.temp = temp.ti;
	imu.tempf = (((float) temp.ti)/340.0) + 35.0;
}

// Yaw function needs work
void Calculate_Roll_Pitch_Yaw(void)
{
	imu.roll = atan2(imu.af[1],imu.af[2])*RADTODEG;
	imu.pitch = atan2(imu.af[0],imu.af[2])*RADTODEG;
//	imu.pitch = atan(-imu.af[0]/sqrt(imu.af[0]*imu.af[0] + imu.af[1]*imu.af[1] + imu.af[2]*imu.af[2]))*RADTODEG;
	imu.yaw = 0.0;  // TBD
}

//#define sampleFreq	512.0f		// sample frequency in Hz
#define sampleFreq	166.0f		// sample frequency in Hz
#define betaDef		0.1f		// 2 * proportional gain

//---------------------------------------------------------------------------------------------------
// Variable definitions

volatile float beta = betaDef;								// 2 * proportional gain (Kp)
volatile float q0 = 1.0f, q1 = 0.0f, q2 = 0.0f, q3 = 0.0f;	// quaternion of sensor frame relative to auxiliary frame

//---------------------------------------------------------------------------------------------------
// Fast inverse square-root
// See: http://en.wikipedia.org/wiki/Fast_inverse_square_root

float invSqrt(float x)
{
	float halfx = 0.5f * x;
	float y = x;

	long i = *(long*)&y;
	i = 0x5f3759df - (i>>1);
	y = *(float*)&i;
	y = y * (1.5f - (halfx * y * y));
	return y;
}

void Init_Quaternion(void)
{
	q0 = 1.0f;
    q1 = 0.0f;
	q2 = 0.0f;
	q3 = 0.0f;
}

#ifdef MADGWICKENABLE
void MadgwickAHRSupdate(float gx, float gy, float gz, float ax, float ay, float az, float mx, float my, float mz)
{
	volatile float recipNorm;
	volatile float s0, s1, s2, s3;
	volatile float qDot1, qDot2, qDot3, qDot4;
	volatile float hx, hy;
	volatile float _2q0mx, _2q0my, _2q0mz, _2q1mx, _2bx, _2bz, _4bx, _4bz, _2q0, _2q1, _2q2, _2q3, _2q0q2, _2q2q3, q0q0, q0q1, q0q2, q0q3, q1q1, q1q2, q1q3, q2q2, q2q3, q3q3;

	// Use IMU algorithm if magnetometer measurement invalid (avoids NaN in magnetometer normalization)
	if((mx == 0.0f) && (my == 0.0f) && (mz == 0.0f)) {
		MadgwickAHRSupdateIMU(gx, gy, gz, ax, ay, az);
		return;
	}

	// Rate of change of quaternion from gyroscope
	qDot1 = 0.5f * (-q1 * gx - q2 * gy - q3 * gz);
	qDot2 = 0.5f * (q0 * gx + q2 * gz - q3 * gy);
	qDot3 = 0.5f * (q0 * gy - q1 * gz + q3 * gx);
	qDot4 = 0.5f * (q0 * gz + q1 * gy - q2 * gx);

	// Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
	if(!((ax == 0.0f) && (ay == 0.0f) && (az == 0.0f))) {

		// Normalise accelerometer measurement
		recipNorm = invSqrt(ax * ax + ay * ay + az * az);
		ax *= recipNorm;
		ay *= recipNorm;
		az *= recipNorm;

		// Normalise magnetometer measurement
		recipNorm = invSqrt(mx * mx + my * my + mz * mz);
		mx *= recipNorm;
		my *= recipNorm;
		mz *= recipNorm;

		// Auxiliary variables to avoid repeated arithmetic
		_2q0mx = 2.0f * q0 * mx;
		_2q0my = 2.0f * q0 * my;
		_2q0mz = 2.0f * q0 * mz;
		_2q1mx = 2.0f * q1 * mx;
		_2q0 = 2.0f * q0;
		_2q1 = 2.0f * q1;
		_2q2 = 2.0f * q2;
		_2q3 = 2.0f * q3;
		_2q0q2 = 2.0f * q0 * q2;
		_2q2q3 = 2.0f * q2 * q3;
		q0q0 = q0 * q0;
		q0q1 = q0 * q1;
		q0q2 = q0 * q2;
		q0q3 = q0 * q3;
		q1q1 = q1 * q1;
		q1q2 = q1 * q2;
		q1q3 = q1 * q3;
		q2q2 = q2 * q2;
		q2q3 = q2 * q3;
		q3q3 = q3 * q3;

		// Reference direction of Earth's magnetic field
		hx = mx * q0q0 - _2q0my * q3 + _2q0mz * q2 + mx * q1q1 + _2q1 * my * q2 + _2q1 * mz * q3 - mx * q2q2 - mx * q3q3;
		hy = _2q0mx * q3 + my * q0q0 - _2q0mz * q1 + _2q1mx * q2 - my * q1q1 + my * q2q2 + _2q2 * mz * q3 - my * q3q3;
		_2bx = sqrt(hx * hx + hy * hy);
		_2bz = -_2q0mx * q2 + _2q0my * q1 + mz * q0q0 + _2q1mx * q3 - mz * q1q1 + _2q2 * my * q3 - mz * q2q2 + mz * q3q3;
		_4bx = 2.0f * _2bx;
		_4bz = 2.0f * _2bz;

		// Gradient decent algorithm corrective step
		s0 = -_2q2 * (2.0f * q1q3 - _2q0q2 - ax) + _2q1 * (2.0f * q0q1 + _2q2q3 - ay) - _2bz * q2 * (_2bx * (0.5f - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) + (-_2bx * q3 + _2bz * q1) * (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) + _2bx * q2 * (_2bx * (q0q2 + q1q3) + _2bz * (0.5f - q1q1 - q2q2) - mz);
		s1 = _2q3 * (2.0f * q1q3 - _2q0q2 - ax) + _2q0 * (2.0f * q0q1 + _2q2q3 - ay) - 4.0f * q1 * (1 - 2.0f * q1q1 - 2.0f * q2q2 - az) + _2bz * q3 * (_2bx * (0.5f - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) + (_2bx * q2 + _2bz * q0) * (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) + (_2bx * q3 - _4bz * q1) * (_2bx * (q0q2 + q1q3) + _2bz * (0.5f - q1q1 - q2q2) - mz);
		s2 = -_2q0 * (2.0f * q1q3 - _2q0q2 - ax) + _2q3 * (2.0f * q0q1 + _2q2q3 - ay) - 4.0f * q2 * (1 - 2.0f * q1q1 - 2.0f * q2q2 - az) + (-_4bx * q2 - _2bz * q0) * (_2bx * (0.5f - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) + (_2bx * q1 + _2bz * q3) * (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) + (_2bx * q0 - _4bz * q2) * (_2bx * (q0q2 + q1q3) + _2bz * (0.5f - q1q1 - q2q2) - mz);
		s3 = _2q1 * (2.0f * q1q3 - _2q0q2 - ax) + _2q2 * (2.0f * q0q1 + _2q2q3 - ay) + (-_4bx * q3 + _2bz * q1) * (_2bx * (0.5f - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) + (-_2bx * q0 + _2bz * q2) * (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) + _2bx * q1 * (_2bx * (q0q2 + q1q3) + _2bz * (0.5f - q1q1 - q2q2) - mz);
		recipNorm = invSqrt(s0 * s0 + s1 * s1 + s2 * s2 + s3 * s3); // normalise step magnitude
		s0 *= recipNorm;
		s1 *= recipNorm;
		s2 *= recipNorm;
		s3 *= recipNorm;

		// Apply feedback step
		qDot1 -= beta * s0;
		qDot2 -= beta * s1;
		qDot3 -= beta * s2;
		qDot4 -= beta * s3;
	}

	// Integrate rate of change of quaternion to yield quaternion
	q0 += qDot1 * (1.0f / sampleFreq);
	q1 += qDot2 * (1.0f / sampleFreq);
	q2 += qDot3 * (1.0f / sampleFreq);
	q3 += qDot4 * (1.0f / sampleFreq);

	// Normalise quaternion
	recipNorm = invSqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3);
	q0 *= recipNorm;
	q1 *= recipNorm;
	q2 *= recipNorm;
	q3 *= recipNorm;
}
//---------------------------------------------------------------------------------------------------
// IMU algorithm update

void MadgwickAHRSupdateIMU(float gx, float gy, float gz, float ax, float ay, float az)
{
	volatile float recipNorm;
	volatile float s0, s1, s2, s3;
	volatile float qDot1, qDot2, qDot3, qDot4;
	volatile float _2q0, _2q1, _2q2, _2q3, _4q0, _4q1, _4q2 ,_8q1, _8q2, q0q0, q1q1, q2q2, q3q3;

	// Rate of change of quaternion from gyroscope
	qDot1 = 0.5f * (-q1 * gx - q2 * gy - q3 * gz);
	qDot2 = 0.5f * (q0 * gx + q2 * gz - q3 * gy);
	qDot3 = 0.5f * (q0 * gy - q1 * gz + q3 * gx);
	qDot4 = 0.5f * (q0 * gz + q1 * gy - q2 * gx);

	// Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
	if(!((ax == 0.0f) && (ay == 0.0f) && (az == 0.0f))) {

		// Normalise accelerometer measurement
		recipNorm = invSqrt(ax * ax + ay * ay + az * az);
		ax *= recipNorm;
		ay *= recipNorm;
		az *= recipNorm;

		// Auxiliary variables to avoid repeated arithmetic
		_2q0 = 2.0f * q0;
		_2q1 = 2.0f * q1;
		_2q2 = 2.0f * q2;
		_2q3 = 2.0f * q3;
		_4q0 = 4.0f * q0;
		_4q1 = 4.0f * q1;
		_4q2 = 4.0f * q2;
		_8q1 = 8.0f * q1;
		_8q2 = 8.0f * q2;
		q0q0 = q0 * q0;
		q1q1 = q1 * q1;
		q2q2 = q2 * q2;
		q3q3 = q3 * q3;

		// Gradient decent algorithm corrective step
		s0 = _4q0 * q2q2 + _2q2 * ax + _4q0 * q1q1 - _2q1 * ay;
		s1 = _4q1 * q3q3 - _2q3 * ax + 4.0f * q0q0 * q1 - _2q0 * ay - _4q1 + _8q1 * q1q1 + _8q1 * q2q2 + _4q1 * az;
		s2 = 4.0f * q0q0 * q2 + _2q0 * ax + _4q2 * q3q3 - _2q3 * ay - _4q2 + _8q2 * q1q1 + _8q2 * q2q2 + _4q2 * az;
		s3 = 4.0f * q1q1 * q3 - _2q1 * ax + 4.0f * q2q2 * q3 - _2q2 * ay;
		recipNorm = invSqrt(s0 * s0 + s1 * s1 + s2 * s2 + s3 * s3); // normalise step magnitude
		s0 *= recipNorm;
		s1 *= recipNorm;
		s2 *= recipNorm;
		s3 *= recipNorm;

		// Apply feedback step
		qDot1 -= beta * s0;
		qDot2 -= beta * s1;
		qDot3 -= beta * s2;
		qDot4 -= beta * s3;
	}

	// Integrate rate of change of quaternion to yield quaternion
	q0 += qDot1 * (1.0f / sampleFreq);
	q1 += qDot2 * (1.0f / sampleFreq);
	q2 += qDot3 * (1.0f / sampleFreq);
	q3 += qDot4 * (1.0f / sampleFreq);

	// Normalise quaternion
	recipNorm = invSqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3);
	q0 *= recipNorm;
	q1 *= recipNorm;
	q2 *= recipNorm;
	q3 *= recipNorm;
}
#endif

void MPU9150_Init(void)
{
	volatile int i;
	
	for(i=0; i<3; i++) {
		imu.ai[i] = imu.gi[i] = imu.mi[i] = 0;
		imu.af[i] = imu.gf[i] = imu.mf[i] = 0.0;
		}
		
	MPU9150_Setup();
	MPU9150_Read_Mag_Sense();  // Access and calculate magnetic sensitivity parameters
	MPU9150_Read_Mag();
	MPU9150_Read_Accel();
	MPU9150_Read_Gyro();
	MPU9150_Read_Temp();
}

void MPU9150_Data_Read(void)
{
	MPU9150_Read_Accel();
	MPU9150_Read_Gyro();
	MPU9150_Read_Mag();
	MPU9150_Read_Temp();	
}

void Get_9DOFMEMsData(volatile char mode)
{
	MPU9150_Data_Read();
}

#ifdef MADGWICKENABLE
void  Quaternion2RollPitchYaw(void)
{
	struct quat {
		float w,x,y,z;
		} q;

	q.x = q0;
	q.y = q1;
	q.z = q2;
	q.w = q3;

	//gx = 2 * (q[1]*q[3] - q[0]*q[2]);
	//gy = 2 * (q[0]*q[1] + q[2]*q[3]);
	//gz = q[0]*q[0] - q[1]*q[1] - q[2]*q[2] + q[3]*q[3];

	//ypr[0] = atan2(2 * q[1] * q[2] - 2 * q[0] * q[3], 2 * q[0]*q[0] + 2 * q[1] * q[1] - 1);
	//ypr[1] = atan(gx / sqrt(gy*gy + gz*gz));
	//ypr[2] = atan(gy / sqrt(gx*gx + gz*gz));

	imu.yaw = atan2(2.0*(q.y*q.z + q.w*q.x), q.w*q.w - q.x*q.x - q.y*q.y + q.z*q.z);
	imu.pitch = asin(-2.0*(q.x*q.z - q.w*q.y));
	imu.roll = atan2(2.0*(q.x*q.y + q.w*q.z), q.w*q.w + q.x*q.x - q.y*q.y - q.z*q.z);
}
#endif

void Get_10DOFMEMsData(volatile char mode)
{
	MPU9150_Data_Read();
#ifdef MADGWICKENABLE
	MadgwickAHRSupdate(imu.gf[0]/RADTODEG,imu.gf[1]/RADTODEG,imu.gf[2]/RADTODEG, imu.af[0],imu.af[1],imu.af[2], imu.mf[0],imu.mf[1],imu.mf[2]);
	Quaternion2RollPitchYaw();
#endif
	//Calculate_Roll_Pitch_Yaw();
	//DisplayBMP180Brief('n');
}

void Update_Display_MPU9150(char mode)
{
	char outstrg[20],outstrg2[25];
	int i;

	for(i=0; i<3; i++) {
		ftoa(imu.af[i],3,7,outstrg);
		Check_RF_Connection();
		if(mode==USBOUT)  Puts(outstrg);
		#ifdef ST7735LCD
			ST7735_fillRect(1+i*41, 91, 41, 10,colorsRGB24toRGB565(255,255,255));
			ST7735_drawString(1+i*41,91,outstrg,colorsRGB24toRGB565(0,0,0),1);
		#endif
		#ifdef  SSD1306OLED
			//sprintf(outstrg,"x=%s",ftoa(imu.af[i],3,7));
			strcpy(outstrg2,"x");
			strcat(outstrg2,outstrg);
			if(i==0) ssd1306_printfs(3,0,outstrg2);
			outstrg2[0] = 'y';
			if(i==1) ssd1306_printfs(4,0,outstrg2);
			outstrg2[0] = 'z';
			if(i==2) ssd1306_printfs(5,0,outstrg2);
		#endif
		}
	for(i=0; i<3; i++) {
		ftoa(imu.gf[i],2,7,outstrg);
		Check_RF_Connection();
		if(mode==USBOUT)  Puts(outstrg);
		#ifdef ST7735LCD
			ST7735_fillRect(1+i*41, 102, 41, 10,colorsRGB24toRGB565(255,255,255));
			ST7735_drawString(1+i*41,102,outstrg,colorsRGB24toRGB565(0,0,0),1);
		#endif
		#ifdef  SSD1306OLED
			if(i==0) ssd1306_printfs(3,8,outstrg);
			if(i==1) ssd1306_printfs(4,8,outstrg);
			if(i==2) ssd1306_printfs(5,8,outstrg);
		#endif
		}
	//if(mode==USBOUT)  printf("%4d%4d%4d  ",imu.mi[0],imu.mi[1],imu.mi[2]);
	Check_RF_Connection();
	#ifdef ST7735LCD
		sprintf(outstrg,"m: %4d  %4d  %4d",imu.mi[0],imu.mi[1],imu.mi[2]);
		ST7735_fillRect(1, 113, 120, 10,colorsRGB24toRGB565(255,255,255));
		ST7735_drawString(1,113,outstrg,colorsRGB24toRGB565(0,0,0),1);
	#endif
	#ifdef  SSD1306OLED
		//sprintf(outstrg,"m:%4d%4d%4d",imu.mi[0],imu.mi[1],imu.mi[2]);
		ssd1306_printfs(6,0,outstrg);
	#endif
	ftoa(imu.roll,2,7,outstrg2);
	strcat(outstrg2," ");
	ftoa(imu.pitch,2,7,outstrg2);
	if(mode==USBOUT)  Puts(outstrg2);
	Check_RF_Connection();
	#ifdef ST7735LCD
		strcpy(outstrg2,"r:");
		strcat(outstrg2,ftoa(imu.roll,2,7));
		strcat(outstrg2," p:");
		strcat(outstrg2,ftoa(imu.pitch,2,7));
		ST7735_fillRect(1, 125, 120, 10,colorsRGB24toRGB565(255,255,255));
		ST7735_drawString(1,125,outstrg2,colorsRGB24toRGB565(0,0,0),1);

		// Dump BMP data to screen every n'th loop to keep processing load down
		if(loopcnt++>5) {
			Check_RF_Connection();
			DisplayBMP180Brief('c');
			loopcnt = 0;
		}
	#endif
}

void MPU9150_Test(void)
{
	volatile char inchar,inchar2,displaymode,linefeed;
	char outstrg[20],outstrg2[25];
	volatile int i,loopcnt=0;
	
	Puts("\n\rMPU-9150 Test\n\r>");
	
	while((inchar=GetCharnw())!='e') {
		switch(inchar) {
			PutChar(inchar);
			case 't':
				MPU9150_Read_Temp();
				//printf("\n\rMPU9150 Temperature=%d ",imu.temp);
				ftoa(imu.tempf,2,7,outstrg);
				Puts(outstrg);
				//printf("compensated=%s\n\r>",ftoa(imu.tempf+mpu9150_tempoffset,2,7));
				break;
			case 's':
				Puts("Read MPU9150 Magnetic sensitivity values\n\r");
				MPU9150_Read_Mag_Sense();
				//printf("asax=%f  asay=%f  asaz=%f\n\r>",imu.magcor[0],imu.magcor[1],imu.magcor[2]);
				break;
			case '8':
				Puts("Reset MPU9150\n\r");
				MPU9150_WriteSensor(MPU9150_PWR_MGMT_1,0x80);
				//printf("pm1=%x\n\r>",MPU9150_ReadSensor2(MPU9150_PWR_MGMT_1));
				break;
			case '9':
				displaymode = 'p';
				linefeed = 0;
				Puts("Get 9 DOF data\n\r");
				MPU9150_Init();
				while((inchar2=GetCharnw())!='e') {
					TIM7_Start();
					Check_RF_Connection();
					if(inchar2=='r') {
						displaymode = inchar2;
						for(i=0; i<79; i++) PutChar(' ');
						PutChar('\r');
						}
					if(inchar2=='p') {
						displaymode = inchar2;
						for(i=0; i<79; i++) PutChar(' ');
						PutChar('\r');
						}
					if(inchar2=='l') linefeed = inchar2;
					if(inchar2=='n') linefeed = 0;
					Get_10DOFMEMsData('r');
					if(displaymode=='r') Puts("data");
						//printf("a=%5d %5d %5d g=%5d %5d %5d m=%4d %4d %4d t=%d \r",imu.ai[0],imu.ai[1],imu.ai[2],
						//													 imu.gi[0],imu.gi[1],imu.gi[2],
						//													 imu.mi[0],imu.mi[1],imu.mi[2],imu.temp);
					else {
						Update_Display_MPU9150(USBOUT);
						}
					TIM7_Stop();
					//printf(" %ld us\r",ElapsedTimeusec());
					//Delaymsec(20);
					if(linefeed)  PutChar('\n');
					}
				Puts("\n\r>");
				break;
			case 'i':
				Puts("Initialize MPU9150\n\r");
				MPU9150_Setup();
				break;
			}
		}
	Puts("\n\r$");
}

